procedure maskstat(image,mask,goodvalue)

# Calculates images statistics from unmasked regions of an input frame.
# * Mask file must have only values of 0 and 1!
# * Parameter 'iterstat' gives the option of calling iterative statistics
#   routine iterstat instead of imstat.   
# * It is assumed that the user will have set the parameters for imstat
#   or iterstat in whatever manner is desired.  This routine does not
#   set those parameters except using lower= to exclude the masked pixels.
# * Checks to see if stsdas.tools is loaded.  If so, it uses imcalc to do the 
#   calculation;  otherwise, it uses a slower combination of imarith statements.
# Calls scripts minv and possibly iterstat.

# **BUG** minmax step means that in an image with a high and nonzero mean,
# 0-valued pixels will not be outside the initial iterstat range, which might 
# make itersig high enough that such pixels will (erroneously, presumably)
# not be excluded in any of the iterations.

# Date: Fri, 21 Oct 94 04:07:39 PDT # From: prme (Peter Eisenhardt)
# Subject: Re:  more on negative scale factors
# 
# iterstat was not converging - not even iterating, actually, leaving me with 
# the initial (negative) median value.  This is due to two problems, really:
# 
# 1) the flag value in maskstat is (minval - 1) which is not in some cases 
# different enough from the lower threshold (minval) sent to iterstat/imstat
# to be noticeable by imstat.  Basically roundoff error, I guess, the
# result of which is that NONE of the "flagged" pixels are excluded by
# the lower threshold, so no convergence from this approach.  
# This has probably been happening frequently, 
# but was not a problem because iterstat picked up the slack.
# 
# 2) The second problem is when the initial sigma becomes so large that 
# mean +/- 5 sigma exceeds the range of data values - thus no
# pixels get rejected in the second iteration.  This happened for
# the first time I think because there are a vast number (1386)
# of bad pixels in the J data, all of which got set to a very low
# flag value in maskstat.  This created a bimodal distribution where sigma 
# is very large and all of the data is within 5 sigma of the mean.  
# 
# A fix which I am implementing is to change the
# flagvalue in maskstat from (minval - 1) to (2*minval - maxval), 
# which ensures that the flagvalue is always well below the range of values
# in the unmasked image, so that the lower threshold in imstat will work.

# 930722: created by Mark Dickinson
# 950216: got rid of imcalc option-- couldn't handle long input image names
# 950314: adjusted use of `sections' task-- risk of collision w/other tasks
# 960405: added option to take masks from a header keyword
# 960413: added option to not print iterstat output
# 960416: fixed bug for 1-entry list, and cleaned up code using `| scan'
# 970429: added insertion of iterstat results into image header if iterstat+
# 970501: fixed bug that pixel masks can't have negative or non-integer values
# 970508: insert iterstat results into image header if iterstat+ and addheader+
# 970620: allow for explicit temporary changing of iterstat params if desired
# 970714: added MASKMASK header param (= mask used) if iterstat+ addheader+
# 970930: fixed bug relating to .pl image being mulitplied by too large a #
	
string	image   	{prompt="Input image(s) or list"}
string	mask    	{prompt="Input mask(s), list, or !<keyword>"}
int	goodvalue	{1,min=0,max=1,prompt="Good pixel value in mask"}

real    nsigrej 	{5.,min=0.,prompt="Number of sigmas for limits"}
int     maxiter 	{10,min=1,prompt="Maximum number of iterations"}
real    lower		{INDEF,prompt="Initial lower limit for data range"}
real    upper		{INDEF,prompt="Initial upper limit for data range"}
bool	verbose 	{no,prompt="Verbose output?"}
bool	iterstat	{no,prompt="Use iterstat instead of imstat?"}
bool	addheader	{yes,prompt="Add iterstat output to image header?"}
bool	printit		{yes,prompt="Print iterstat output?"}
struct	*inimglist
struct	*mlist

begin

	string  check		# `@filename' check
	string	imglist		# equals image
	string	ext		# equals image extension
	string	msk		# equals mask
	string	mskfile		# temporary mask file
	string	keyword		# mask header keyword
	string	mfile		# name of individual mask file
	string	mf		# name of individual mask file without .pl 
	string	minvfile	# inverted mask file
	string  infile          # temporary list for files
	string  j4		# temp string
	int	goodv		# equals goodvalue
	int	nimgs,nmask	# number of input images and masks
	int	j1,j2,j3 	# temp values
	real	minval		# lower limit for image statistics calculation
	real	maxval		# upper limit of image statistics 
	real	flagval		# flag value
	bool    indel           # delete input list file?
	bool    maskdel         # delete mask list file?

# Get query parameters & expand input lists
	imglist = image
	msk = mask
                if(msk=="")msk="!BPM"
	goodv = goodvalue
	maskdel = no
	indel = no

# Expand file lists into temporary files.
	infile = imglist
	check = substr(imglist,1,1)
        if(check=="@"){
		infile=substr(imglist,2,strlen(imglist))
		count(infile) | scan(nimgs)
	}
	else {
		infile = mktemp("tmp$maskstat")
		indel=yes
		sections(imglist,option="fullname",>infile)
		count(infile) | scan(nimgs)
	}
	inimglist = infile

	#print("one")

# Exact same spiel for list of masks, with extra step if only one mask.
	mskfile = msk
	check = substr(msk,1,1)
	if(check=="@" || check=="!"){
		if(check=="!"){
			maskdel=yes
			keyword=substr(msk,2,strlen(msk))
			mskfile = mktemp("tmp$maskstat")
			hsel("@"//infile,(keyword),yes, >> mskfile)
		}
		if(check=="@"){
			mskfile=substr(msk,2,strlen(msk))
		}
		count (mskfile) | scan (nmask)
	}
	else {
		mskfile = mktemp("tmp$maskstat")
		maskdel=yes
		sections(msk,option="fullname",>mskfile)
		count (mskfile) | scan (nmask)
		if (nmask == 1) {
			mfile = msk
			fileroot(mfile,validim+)
			minvfile = "_"//fileroot.root//".pl"
			if(access(minvfile))imdel(minvfile,ver-)
			minv (mfile,minvfile)
		}
	}
	mlist = mskfile
	if (nmask != nimgs && nmask != 1) {
		print ("ERROR: Numbers of image and mask files differ.")
		return
	}

	#print("two")

# Loop through input files.

	while (fscan(inimglist,img) != EOF) {

		fileroot(img,validim+)
		ext=fileroot.extension
		if(ext!="")ext="."//ext
if(access("_"//fileroot.root//ext))imdel("_"//fileroot.root//ext,ver-)
if(access("_"//fileroot.root//".imh"))imdel("_"//fileroot.root//".imh",ver-)
if(access("_"//fileroot.root//".hhh"))imdel("_"//fileroot.root//".hhh",ver-)

	#print("thr")

# Get name of mask file 

#		if (check == "@" || check == "!") {
			if (fscan(mlist,mfile) != EOF) {
				fileroot(mfile,validim+)
#				minvfile = "_"//fileroot.root//".pl"
				minvfile = "_"//fileroot.root//".imh"
				if(access(minvfile))imdel(minvfile,ver-)
				minv (mfile,minvfile)
			   } else {
				print ("ERROR: Premature end of mask list.")
				return
			}
#		}

	#print("for")

# Strip extension off mask name and save root as variable mf for later use.

		fileroot (mfile,validim+)
		mf = fileroot.root
                ext=fileroot.extension
                if(ext!="")ext="."//ext
		if(access("_x"//mf//ext))imdel("_x"//mf//ext,ver-)
		if(access("_x"//mf//".imh"))imdel("_x"//mf//".imh",ver-)
		if(access("_x"//mf//".hhh"))imdel("_x"//mf//".hhh",ver-)

	#print("fiv")

if (verbose) print ("Calculating statistics for ",img," using mask ",mfile)

		minmax (img,force+,update-,ver-)
		minval = minmax.minval 
		maxval = minmax.maxval 
		if (lower == INDEF) {
			flagval = 2 * minval - maxval
		   } else {
			minval  = lower
			flagval = 2 * lower - maxval
		}

	#print("six")
	#print(img//" "//mfile//" "//minvfile)

# Masking replaces bad pixels with "flagval" and leaves good pixels unchanged.
		   if (goodv == 0) {
imar (img,"*",minvfile,"_"//img,pixty="real",calcty="real",ver-)
imar (mfile,"*",abs(int(flagval)),"_x"//mf,pixty="real",calcty="real",ver-)
imar ("_"//img,"-","_x"//mf,"_"//img,pixty="real",calcty="real",ver-)
		      } else {
imar (img,"*",mfile,"_"//img,pixty="real",calcty="real",ver-)
imar (minvfile,"*",abs(int(flagval)),"_x"//mf,pixty="real",calcty="real",ver-)
imar ("_"//img,"-","_x"//mf,"_"//img,pixty="real",calcty="real",ver-)
		   }

	#print("sev")

		if (iterstat) {
			if (printit) {
				iterstat ("_"//img,lower=minval,nsigrej=nsigrej,maxiter=maxiter,upper=upper,verbose=verbose,print+,addh-)
			} else {
				iterstat ("_"//img,lower=minval,nsigrej=nsigrej,maxiter=maxiter,upper=upper,verbose=verbose,print-,addh-)
			}
			if (addheader) {
		hedit (img, "MASKMASK", msk, add+, ver-,show-,update+)
		hedit (img, "ITERMEAN", iterstat.mean, add+, ver-,show-,update+)
                hedit (img, "ITERSIG", iterstat.sigma, add+, ver-,show-,update+)
                hedit (img, "VARIANCE", iterstat.variance,add+,ver-,sho-,updat+)
                hedit (img, "ITERMED", iterstat.median, add+,ver-,show-,update+)
                hedit (img, "ITERMODE", iterstat.valmode,add+,ver-,show-,updat+)
			}
		} else {
	imstat   ("_"//img,lower=minval,upper=upper)
		}

		imdelete ("_"//img,ver-)
		imdelete ("_x"//mf,ver-)
		if (nmask != 1) imdelete (minvfile,ver-)

	}

	#print("ate")

	if (nmask == 1) imdelete (minvfile,ver-)

	#print("nin")

        if(indel)delete(infile,ver-,>&"dev$null")
        if(maskdel)delete(mskfile,ver-,>&"dev$null")
	inimglist = ""
	mlist = ""

end	

